#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>

#include <termios.h>
#include <fcntl.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>

#include "localserver.h"
#include "blueserver.h"

#define BT_HID_REMOTE

//#define DEBUG_BLUESERVER

/**
 * The defines USE_FIXED_ADDR and FIXED_ADDR_FILE are used for test purposes
 * to limit the device-discovery to a single, predetermined remote device.
 * Main reason is to speedup discovery.
 * Enable/disable USE_FIXED_ADDR to enable/disable this feature
 */
//#define USE_FIXED_ADDR
#define FIXED_ADDR_FILE	"/mnt/flash/fixed_bt_addr"

// Taken from FRAMECOMMON/linux/nonbluez.h
struct THCIDeviceRequirements
{
  unsigned short iDeviceID;
  unsigned long  iDeviceOptions;
};

// Taken from GameFrame\bluetooth\btchannel.h
#define BluetoothProtocolHCI    1
#define HciDeviceDisableScan    0
#define HciDeviceScanInquiry    1
#define HciDeviceScanPage       2


#ifdef DEBUG_BLUESERVER
  #define LS_DEBUG(msg...)     do {          printf(msg); fflush(stdout); }   while (0)
  #define LS_DEBUGT(t, msg...) do { if (t) { printf(msg); fflush(stdout); } } while (0)
#else
  #define LS_DEBUG(msg...)     do { } while (0)
  #define LS_DEBUGT(t, msg...) do { } while (0)
#endif


// major class of device codes
#define MAJ_COD_MISC                  0   // 00000 Miscellaneous
#define MAJ_COD_COMPUTER              1   // 00001 Computer
#define MAJ_COD_PHONE                 2   // 00010 Phone
#define MAJ_COD_LAN                   3   // 00011 LAN / Network Access point
#define MAJ_COD_AUDIO_VIDEO           4   // 00100 Audio / Video
#define MAJ_COD_COMPUTER_PERIPHERAL   5   // 00101 Peripheral
#define MAJ_COD_IMAGING               6   // 00110 Imaging
#define MAJ_COD_WEARABLE              7   // 00111 Wearable
#define MAJ_COD_TOY                   8   // 01000 Toy
#define MAJ_COD_MEDICAL               9   // 01001 Medical
#define MAJ_COD_UNCATEGORIZED        31   // 11111 Uncategorized


// minor class of device types (where MajorDeviceType is Computer or PDA)
#define MAJ_COMP_MIN_LAPTOP           0x03  // Laptop
#define MAJ_COMP_MIN_HANDHELD         0x04  // Handheld PC/PDA (clam shell)
#define MAJ_COMP_MIN_PALMSIZED        0x05  // Palm sized PC/PDA

// minor class of device types (where MajorDeviceType is Audio or Video device)
#define MAJ_AV_MIN_HEADSETDEVICE      0x01  // Wearable Headset Device
#define MAJ_AV_MIN_HANDSFREE          0x02  // Handsfree device
#define MAJ_AV_MIN_HEADSET_HANDSFREE  0x03  // Handsfree and Headset device; actually Reserved
#define MAJ_AV_MIN_LOUDSPEAKER        0x05  // Loudspeaker
#define MAJ_AV_MIN_HEADPHONE          0x06  // A2DP headphones
#define MAJ_AV_MIN_PORTABLE           0x07  // Portable audio
#define MAJ_AV_MIN_CARAUDIO           0x08  // Car audio
#define MAJ_AV_MIN_HIFIDEVICE         0x0A  // HiFi device
#define MAJ_AV_MIN_VIDEO_AND_SP       0x0F  // Video Display and Loudspeaker

// service classes
#define COS_OBEXTRANSFER             (1L) << 20 // Obex Transfer service class

#define DEBUG_CLASS_OF_DEVICE_FILTERING

char outbuf[1024];

int scan_profiles_fd = -1;
typedef int (*handler_t)(bdaddr_t *bdaddr, struct search_context *arg);
static bdaddr_t interface_bdaddr;
int estr2ba(const char *str, bdaddr_t *ba)
{
  // Only trap "local", "any" is already dealt with
  if(!strcmp(str, "local"))
  {
    bacpy(ba, BDADDR_LOCAL);
    return 0;
  }
  return str2ba(str, ba);
}

// Pass args to the inquiry/search handler
struct search_context
{
  char      *svc;    // Service
  uuid_t    group;   // Browse group
  int       tree;    // Display full attribute tree
  uint32_t  handle;  // Service record handle
};



static char UUID_str[MAX_LEN_UUID_STR];

static void writestr(int fd, const char *str)
{
write_retry:
  size_t len = strlen(str) + 1; // add NUL terminator
  ssize_t nw = write(fd, str, len);
  if (nw < 0)
  {
    if (errno == EINTR)
      goto write_retry;
    printf("blueserver: unable to write to socket: %s\n", strerror(errno));
  }
  else if (static_cast<size_t>(nw) != len)
  {
    printf("blueserver: only wrote %zd of %zu bytes to socket\n", nw, len);
  }
}


static void scan_profiles_print_service_class(void *value, void *userData)
{
  LS_DEBUG("~~~~~~~ scan_profiles_print_service_class called\n");
  char ServiceClassUUID_str[MAX_LEN_SERVICECLASS_UUID_STR];
  uuid_t *uuid = (uuid_t *)value;

  sdp_uuid2strn(uuid, UUID_str, MAX_LEN_UUID_STR);
  sdp_svclass_uuid2strn(uuid, ServiceClassUUID_str, MAX_LEN_SERVICECLASS_UUID_STR);
  char dummy[128];
  sprintf(dummy,"0x%s|%s|", UUID_str, ServiceClassUUID_str);
  strcat(outbuf,dummy);
}


static void scan_profiles_print_service_desc(void *value, void *user)
{
  LS_DEBUG("~~~~~~~ scan_profiles_print_service_desc called\n");
  char str[MAX_LEN_PROTOCOL_UUID_STR];
  sdp_data_t *p = (sdp_data_t *)value;
  int i = 0, proto = 0;

  for (; p; p = p->next, i++)
  {
    switch (p->dtd)
    {
    case SDP_UUID16:
    case SDP_UUID32:
    case SDP_UUID128:
      sdp_uuid2strn(&p->val.uuid, UUID_str, MAX_LEN_UUID_STR);
      sdp_proto_uuid2strn(&p->val.uuid, str, sizeof(str));
      proto = sdp_uuid_to_proto(&p->val.uuid);
      break;
    case SDP_UINT8:
      if (proto == RFCOMM_UUID)
      {
        sprintf(outbuf,"Channel|%d|\n", p->val.uint8);
        writestr(scan_profiles_fd,outbuf);
      }
      break;
    default:
      break;
    }
  }
}


void scan_profiles_print_access_protos(void* vr, void* userData)
{
  sdp_list_t *value = (sdp_list_t *)vr;
  sdp_list_t *protDescSeq = (sdp_list_t *)value;
  sdp_list_foreach(protDescSeq, scan_profiles_print_service_desc, 0);
}


//
// Parse a SDP record in user friendly form.
//
void scan_profiles_print_service_attr(sdp_record_t *rec)
{
  LS_DEBUG("~~~~~~~ scan_profiles_print_service_attr called\n");
  sdp_list_t *list = 0, *proto = 0;

  sdp_data_t *d = sdp_data_get(rec, SDP_ATTR_SVCNAME_PRIMARY);
  if (d)
  {
    sprintf(outbuf,"Name|%s|\n", (d->val.str ? d->val.str : ""));
    writestr(scan_profiles_fd,outbuf);
  }
  else
  {
    d = sdp_data_get(rec, SDP_ATTR_SVCDESC_PRIMARY);
    if (d)
    {
      sprintf(outbuf,"Name|%s|\n", (d->val.str ? d->val.str : ""));
      writestr(scan_profiles_fd,outbuf);
    }
    else
    {
      d = sdp_data_get(rec, SDP_ATTR_PROVNAME_PRIMARY);
      if (d)
      {
        sprintf(outbuf,"Name|%s|\n", (d->val.str ? d->val.str : ""));
        writestr(scan_profiles_fd,outbuf);
      }
    }
  }

//  printf("Service RecHandle: 0x%x\n", rec->handle);

  if (sdp_get_service_classes(rec, &list) == 0)
  {
//    printf("Service Class ID List:\n");
    sprintf(outbuf,"ServiceClass|");
    sdp_list_foreach(list, scan_profiles_print_service_class, 0);
    strcat(outbuf,"\n");
    writestr(scan_profiles_fd,outbuf);
    sdp_list_free(list, free);
  }
  if (sdp_get_access_protos(rec, &proto) == 0)
  {
//    printf("Protocol Descriptor List:\n");
    sdp_list_foreach(proto, scan_profiles_print_access_protos, 0);
    sdp_list_free(proto, (sdp_free_func_t)sdp_data_free);
  }
}


//
// Perform an inquiry and search/browse all peer found.
//
static void scan_profiles_inquiry(handler_t handler, void *arg)
{
  LS_DEBUG("~~~~~~~ scan_profiles_inquiry\n");
  inquiry_info ii[20];
  uint8_t count = 0;
  int i;

  LS_DEBUG("blueserver: Inquiring ...\n");
  if (sdp_general_inquiry(ii, 20, 8, &count) < 0)
  {
    LS_DEBUG("blueserver: Inquiry failed\n");
    return;
  }
  for (i=0; i<count; i++)
    handler((bdaddr_t *)&ii[i].bdaddr, (struct search_context *)arg);
}

/*
 * Search for a specific SDP service
 */
int scan_profiles_do_search(bdaddr_t *bdaddr, struct search_context *context)
{
  LS_DEBUG("~~~~~~~ scan_profiles_do_search called\n");
  sdp_list_t *attrid, *search, *seq, *next;
  uint32_t range = 0x0000ffff;
  char str[20];
  sdp_session_t *sess;

  if (!bdaddr)
  {
    scan_profiles_inquiry(scan_profiles_do_search, context);
    return 0;
  }
  sess = sdp_connect(&interface_bdaddr, bdaddr, SDP_RETRY_IF_BUSY);
  ba2str(bdaddr, str);
  if (!sess) {
    LS_DEBUG("blueserver: Failed to connect to SDP server on %s: %s\n", str, strerror(errno));
    return -1;
  }
  LS_DEBUG("blueserver: sdp_connect call succeded!\n");

  attrid = sdp_list_append(0, &range);
  search = sdp_list_append(0, &context->group);
  if (sdp_service_search_attr_req(sess, search, SDP_ATTR_REQ_RANGE, attrid, &seq)) {
    LS_DEBUG("blueserver: Service Search failed: %s\n", strerror(errno));
    sdp_close(sess);
    return -1;
  }

  LS_DEBUG("blueserver: sdp_service_search_attr_req call succeded!\n");

  sdp_list_free(attrid, 0);
  sdp_list_free(search, 0);

  for (; seq; seq = next)
  {
    sdp_record_t *rec = (sdp_record_t *) seq->data;
    struct search_context sub_context;

    if (context->tree)
    {
      /* Display full tree */
//      sdp_printf_service_attr(rec);
    }
    else
    {
      /* Display user friendly form */
      scan_profiles_print_service_attr(rec);
    }
//    printf("\n");

    if (sdp_get_group_id(rec, &sub_context.group) != -1)
    {
      /* Set the subcontext for browsing the sub tree */
      memcpy(&sub_context, context, sizeof(struct search_context));
      /* Browse the next level down if not done */
      if (sub_context.group.value.uuid16 != context->group.value.uuid16)
        scan_profiles_do_search(bdaddr, &sub_context);
    }
    next = seq->next;
    free(seq);
    sdp_record_free(rec);
  }
  sdp_close(sess);
//  printf("END\n");
  return 0;
}

/*
 * Browse the full SDP database (i.e. list all services starting from the
 * root/top-level).
 */
void scan_profiles_cmd_browse(const char* address)
{
  bacpy(&interface_bdaddr, BDADDR_ANY);

  struct search_context context;
  bdaddr_t bdaddr;

  estr2ba(address, &bdaddr);

  /* Initialise context */
  memset(&context, '\0', sizeof(struct search_context));
#if 0
  sdp_uuid16_create(&(context.group), PUBLIC_BROWSE_GROUP);
  if (scan_profiles_do_search(&bdaddr, &context) != -1)
	return;
  LS_DEBUG("Scanning PUBLIC_BROWSE_GROUP failed!\n");
#else
  sdp_uuid16_create(&(context.group), L2CAP_UUID);
  if (scan_profiles_do_search(&bdaddr, &context) != -1)
	return;
  LS_DEBUG("Scanning L2CAP_UUID failed!\n");
#endif
  /* Scan for specific profiles if general scan fails */
  sdp_uuid16_create(&(context.group), HANDSFREE_SVCLASS_ID);
  if (scan_profiles_do_search(&bdaddr, &context)  == -1)
    LS_DEBUG("Scanning HANDSFREE_SVCLASS_ID failed!\n");
  sdp_uuid16_create(&(context.group), DIALUP_NET_SVCLASS_ID);
  if (scan_profiles_do_search(&bdaddr, &context) == -1)
    LS_DEBUG("Scanning DIALUP_NET_SVCLASS_ID failed!\n");
  sdp_uuid16_create(&(context.group), HEADSET_SVCLASS_ID);
  if (scan_profiles_do_search(&bdaddr, &context) == -1)
    LS_DEBUG("Scanning HEADSET_SVCLASS_ID failed!\n");
  sdp_uuid16_create(&(context.group), AUDIO_SINK_SVCLASS_ID);
  if (scan_profiles_do_search(&bdaddr, &context) == -1)
    LS_DEBUG("Scanning AUDIO_SINK_SVCLASS_ID failed!\n");
  sdp_uuid16_create(&(context.group), SERIAL_PORT_SVCLASS_ID);
  if (scan_profiles_do_search(&bdaddr, &context) == -1)
    LS_DEBUG("Scanning SERIAL_PORT_SVCLASS_ID failed!\n");

  return;
}

#define CSR_VARID_WARM_HALT 0x4004

const char *bccmd_err_names[] = {
  "ok",
  "no such variable id",
  "value too big",
  "no value",
  "bad request",
  "no access",
  "read only",
  "write only",
  "error",
  "permission denied",
};

int csr_write_varid_uint16(int dd, uint16_t seqnum, uint16_t varid, uint16_t value)
{
  unsigned char cmd[] = {
    0x02, 0x00,
    0x09, 0x00,
    seqnum & 0xff, seqnum >> 8,
    varid & 0xff, varid >> 8,
    0x00, 0x00,      // status
    0x00, 0x00,      // payload
    0x00, 0x00,
    0x00, 0x00,
    0x00, 0x00
  };

  unsigned char cp[254], rp[254];
  struct hci_request rq;
  uint16_t status;

  memset(&cp, 0x00, sizeof(cp));
  cp[0] = 0xc2;
  memcpy(cp + 1, cmd, sizeof(cmd));

  memset(&rq, 0, sizeof(rq));
  rq.ogf    = OGF_VENDOR_CMD;
  rq.ocf    = 0x00;
  rq.event  = EVT_VENDOR;
  rq.cparam = cp;
  rq.clen   = sizeof(cmd) + 1;
  rq.rparam = rp;
  rq.rlen   = sizeof(rp);

  if(hci_send_req(dd, &rq, 2000) < 0)
  {
    LS_DEBUG("blueserver: hci_send_req %d %s\n", errno, strerror(errno));
    return -1;
  }

  if(rp[0] != 0xc2)
  {
    LS_DEBUG("blueserver: hci_send_req ret %d %s\n", errno, strerror(errno));
    errno = EIO;
    return -1;
  }

  status = rp[9] + (rp[10] << 8);
  if(status != 0)
  {
    LS_DEBUG("blueserver: Illegal request: %s\n", bccmd_err_names[status]);
    return -1;
  }

  return 0;
}

void EnableBluetooth(int aEnable)
{
  static uint16_t seq = 1;

  LS_DEBUG("blueserver: EnableBluetooth called with %d\n",aEnable);

  if (!aEnable)
  {
    int dev_id = hci_devid("hci0");
    if(dev_id < 0)
    {
      LS_DEBUG("blueserver: Invalid device\n");
      return;
    }
    int dd = hci_open_dev(dev_id);
    if(dd < 0)
    {
      LS_DEBUG("blueserver: Error opening hci device\n");
      return;
    }
    // Highly probable that the GETRESP for this command will not reach the caller.
    csr_write_varid_uint16(dd, seq++, CSR_VARID_WARM_HALT, 0);
    hci_close_dev(dd);
  }
  else
  {
    // Enabling done in script only.
  }
}

bool deviceIsATomTom(bdaddr_t * macid)
{
  // Check on Bluetooth Address Range : 
  // 00:13:6c:00:00:00 - 00:13:6c:ff:ff:ff  
  bool validFirstRange  = (macid->b[5] == 0x00) && 
                          (macid->b[4] == 0x13) &&
                          (macid->b[3] == 0x6c);
  // 00:21:3e:00:00:00 - 00:21:3e:ff:ff:ff
  bool validSecondRange = (macid->b[5] == 0x00) && 
                          (macid->b[4] == 0x21) &&
                          (macid->b[3] == 0x3e);

  return (validFirstRange || validSecondRange);
}

int IsHeadSet(int majorClassOfDevice, int minorClassOfDevice)
{
  return (majorClassOfDevice == MAJ_COD_AUDIO_VIDEO &&
          minorClassOfDevice == MAJ_AV_MIN_HEADSETDEVICE);
}

int IsA2dp(int majorClassOfDevice, int minorClassOfDevice)
{
  return (majorClassOfDevice == MAJ_COD_AUDIO_VIDEO &&
          (minorClassOfDevice == MAJ_AV_MIN_HEADSETDEVICE ||
           minorClassOfDevice == MAJ_AV_MIN_HEADPHONE     ||
           minorClassOfDevice == MAJ_AV_MIN_LOUDSPEAKER   ||
           minorClassOfDevice == MAJ_AV_MIN_PORTABLE      ||
           minorClassOfDevice == MAJ_AV_MIN_CARAUDIO      ||
           minorClassOfDevice == MAJ_AV_MIN_HIFIDEVICE    ||
           minorClassOfDevice == MAJ_AV_MIN_VIDEO_AND_SP  ||
           minorClassOfDevice == MAJ_AV_MIN_HEADSET_HANDSFREE ||
           minorClassOfDevice == MAJ_AV_MIN_HANDSFREE));
}

int IsPhone(int majorClassOfDevice)
{
  return (majorClassOfDevice == MAJ_COD_PHONE);
}

int IsComp(int majorClassOfDevice, int minorClassOfDevice)
{
  return (majorClassOfDevice == MAJ_COD_COMPUTER &&
          (minorClassOfDevice == MAJ_COMP_MIN_HANDHELD ||
           //minorClassOfDevice == MAJ_COMP_MIN_LAPTOP ||
           minorClassOfDevice == MAJ_COMP_MIN_PALMSIZED)) ;
}
int IsOBEX(int classOfService)// e.g for other GO devices
{
  return classOfService & COS_OBEXTRANSFER;
}

int IsBremen(int majorClassOfDevice)
{
  return (majorClassOfDevice == MAJ_COD_LAN );
}

static void requesthandler(const char* request, int fd)
{
  if (!strncmp(request,"DISCOVER",8))
  {
    int bluetooth_discovery = (strlen(request)==11 && !strncmp(request, "DISCOVER BT",11));
    int headset_discovery = (strlen(request)==11 && !strncmp(request,"DISCOVER HS",11));
    int a2dp_discovery = (strlen(request)==13 && !strncmp(request,"DISCOVER A2DP",13));
    int obex_discovery = (strlen(request)==strlen("DISCOVER OBEX") && !strncmp(request,"DISCOVER OBEX",strlen("DISCOVER OBEX")));
    LS_DEBUGT(bluetooth_discovery, "blueserver: bluetooth discovery\n");
    LS_DEBUGT(headset_discovery, "blueserver: headset discovery\n");
    LS_DEBUGT(a2dp_discovery, "blueserver: stereo headphones discovery (A2DP)\n");
    LS_DEBUGT(obex_discovery, "blueserver: OBEX device discovery\n");

    inquiry_info *info = NULL;
    int num_rsp, length, flags;
    char addr[18];
    char name[248];
    char type[8];
    int i, dd;

    length  = 8;  /* ~10 seconds */
    num_rsp = 100;
    flags = IREQ_CACHE_FLUSH;
    int dev_id = -1;
    if (dev_id < 0)
    {
      dev_id = hci_get_route(NULL);
      if (dev_id < 0)
      {
        perror("Device is not available");
        writestr(fd, "END");
        return;
      }
    }

    LS_DEBUG("blueserver: Scanning ...\n");
    int check_count = 0;
REDO:
    num_rsp = hci_inquiry(dev_id, length, num_rsp, NULL, &info, flags);
    if (num_rsp == -1 && check_count < 30)  // 30 sec retry period
    {
      LS_DEBUG("blueserver: RETRYING  num_rsp=%d  check_count=%d\n", num_rsp, check_count);
      sleep(1);
      check_count++;
      goto REDO;
    }

    LS_DEBUG("blueserver: num_rsp = %d\n",num_rsp);

    if (num_rsp >= 0)
    {
       dd = hci_open_dev(dev_id);
      if (dd >= 0)
      {

#ifdef USE_FIXED_ADDR
        struct fixed_addr_t {
          char addr[18];
          bool ok;
        } fixed_addr;
        memset(fixed_addr.addr, 0, sizeof(fixed_addr.addr));

        fixed_addr.ok = false;

        if (FILE *f = fopen(FIXED_ADDR_FILE, "r"))
        {
          size_t sz = fread(&fixed_addr.addr, 1, 17, f);
          fclose(f);

          if (sz<17)
            LS_DEBUG("blueserver: fixed_addr: Did not read enough chars for mac address: %s\n", fixed_addr.addr);
          else
            fixed_addr.ok = true;
        }
        else
        {
          LS_DEBUG("blueserver: fixed_addr: Could not open file %s\n", FIXED_ADDR_FILE);
        }

        LS_DEBUG("blueserver: fixed_addr: fixed_addr.addr=%s, fixed_addr.ok=%d\n", fixed_addr.addr, fixed_addr.ok);
#endif // USE_FIXED_ADDR

        for (i = 0; i < num_rsp; i++)
        {
          int majorClassOfDevice =  (((info+i)->dev_class[1])&0x1F);       // 5 bits
          int minorClassOfDevice = ((((info+i)->dev_class[0])>>2)&0x3F);   // 6 bits
          unsigned long classOfService = ((unsigned long)((info+i)->dev_class[2]) << 16) + ((unsigned long)((info+i)->dev_class[1]) << 8) + (unsigned long)(info+i)->dev_class[0];

          #ifdef DEBUG_CLASS_OF_DEVICE_FILTERING
          ba2str(&(info+i)->bdaddr, addr);
          LS_DEBUG("blueserver [%s]: COS=0x%lX | majCOD=%d | minCOD=%d | ", addr, classOfService, majorClassOfDevice, minorClassOfDevice);
          switch (majorClassOfDevice)
          {
            case MAJ_COD_MISC:                LS_DEBUG("Miscellaneous\n"); break;
            case MAJ_COD_COMPUTER:            LS_DEBUG("Computer (desktop,notebook, PDA, organizers)\n"); break;
            case MAJ_COD_PHONE:               LS_DEBUG("Phone (cellular, cordless, payphone, modem)\n"); break;
            case MAJ_COD_LAN:                 LS_DEBUG("LAN/Network Access point\n"); break;
            case MAJ_COD_AUDIO_VIDEO:         LS_DEBUG("Audio/Video (headset,speaker,stereo, video display, vcr)\n"); break;
            case MAJ_COD_COMPUTER_PERIPHERAL: LS_DEBUG("Peripheral (mouse, joystick, keyboards)\n"); break;
            case MAJ_COD_IMAGING:             LS_DEBUG("Imaging (printing, scanner, camera, display)\n"); break;
            case MAJ_COD_WEARABLE:            LS_DEBUG("Wearable\n"); break;
            case MAJ_COD_TOY:                 LS_DEBUG("Toy\n"); break;
            case MAJ_COD_MEDICAL:             LS_DEBUG("Medical\n"); break;
            case MAJ_COD_UNCATEGORIZED:       LS_DEBUG("Uncategorized, specific device code not specified\n"); break;
            default:                          LS_DEBUG("Unknown class of device\n"); break;
          }
          #endif

          int skipType = 1;
          int deviceType = 0;

          if ( bluetooth_discovery ) //search for all supported BT devices
          {
            if( IsHeadSet(majorClassOfDevice, minorClassOfDevice ) ||
                IsA2dp(majorClassOfDevice, minorClassOfDevice) ||
                IsPhone(majorClassOfDevice) ||
                IsOBEX(classOfService) ||
                IsComp(majorClassOfDevice, minorClassOfDevice) ||
                IsBremen(majorClassOfDevice) )
                    skipType = 0;
          }
          else
          if (headset_discovery)
          {
            if ( IsHeadSet(majorClassOfDevice, minorClassOfDevice ) )
              skipType = 0;
          }
          else if (a2dp_discovery)
          {
            if (IsA2dp(majorClassOfDevice, minorClassOfDevice))
              skipType = 0;
          }
          else if (obex_discovery) //this discover all devices user can send the file to
          {
            if( IsOBEX(classOfService) )
              skipType = 0;
          }
          else // phones discovery
          {
            if (IsPhone(majorClassOfDevice) ||
                IsComp(majorClassOfDevice, minorClassOfDevice))
              skipType = 0;
          }

#ifdef USE_FIXED_ADDR
        if (fixed_addr.ok)
          if (!strncmp(addr, fixed_addr.addr,17))
            skipType = 0;  // this is the one we are looking for: process it
          else
          {
            skipType = 1;  // this is NOT the one we are looking for: skip it
            LS_DEBUG("blueserver: fixed_addr: Skipping address [%s] (not matching fixed addr [%s])\n",addr, fixed_addr.addr);
          }
#endif // USE_FIXED_ADDR

          if (skipType)
          {
            #ifdef DEBUG_CLASS_OF_DEVICE_FILTERING
            LS_DEBUG("blueserver: Skipping (not a phone/PDA, headset or HiFi or OBEX capable device)\n");
            #endif

            continue;
          }

          // If the client sends *any* data, just stop scanning
          {
            int nread;
            ioctl(fd, FIONREAD, &nread);
            if (nread>0)
            {
              LS_DEBUG("blueserver: Stopping scanning for bluetooth devices\n");
              break;
            }
          }


          memset(name, 0, sizeof(name));
          if (hci_read_remote_name(dd, &(info+i)->bdaddr, sizeof(name), name, 100000) < 0)
          {
            #ifdef DEBUG_CLASS_OF_DEVICE_FILTERING
            LS_DEBUG("blueserver: Unable to read device name: hci_read_remote_name failed\n");
            #endif

            // Skip devices with no name
//            strcpy(name, "n/a");
            continue;
          }
          ba2str(&(info+i)->bdaddr, addr);

          if (majorClassOfDevice == MAJ_COD_AUDIO_VIDEO)
          {
            if (deviceIsATomTom(&(info+i)->bdaddr))
            {
              deviceType = BT_DEVICE_TYPE_TOMTOM;
            }
            else
            {
              deviceType = BT_DEVICE_TYPE_GENERIC;
            }
          }
          else if (majorClassOfDevice == MAJ_COD_COMPUTER)
          {
            deviceType = BT_DEVICE_TYPE_GENERIC;
          }
          else if (majorClassOfDevice == MAJ_COD_PHONE)
          {
            deviceType = BT_DEVICE_TYPE_PHONE;
          }
          else if ( ( majorClassOfDevice == MAJ_COD_LAN ) &&
                    ( deviceIsATomTom(&(info+i)->bdaddr)) &&
                    ( !strncmp(name,"LINK",4) ) )
          {
              deviceType = BT_DEVICE_TYPE_BREMEN;
          }
          else deviceType = BT_DEVICE_TYPE_UNKNOWN;

          memset(type, 0, sizeof(type));
          snprintf(type, sizeof(type), "%d", deviceType);
          char buf[256];
          snprintf(buf,256,"%s|%s|%s", addr, name, type);
          writestr(fd, buf);

          #ifdef DEBUG_CLASS_OF_DEVICE_FILTERING
          LS_DEBUG("blueserver: name = %s\n", buf);
          #endif

#ifdef USE_FIXED_ADDR
          if (fixed_addr.ok && !strncmp(addr, fixed_addr.addr,17))
          {
            LS_DEBUG("blueserver: fixed_addr: Found fixed address %s, get out\n", addr);
            break;
          }
#endif // USE_FIXED_ADDR

        }
        close(dd);
        free(info);
      }
    }
    writestr(fd, "END");
  }
  else if (!strncmp(request,"PROFILES",8))
  {
    #define BT_CLIENT_PINFILE   "/etc/bluetooth/client_pin"

    // "PROFILE bdaddr" where bdaddr is macid of client
    // "PROFILE [HS|HF]|pin|bdaddr" where pin is the client pincode to use on bdaddr
    int profiles_a2dp = (strlen(request)>=14 && !strncmp(request,"PROFILES A2DP|",14));
    int profiles_headset = (strlen(request)>=12 && !strncmp(request,"PROFILES HS|",12));
    int profiles_protect = (strlen(request)>=12 && !strncmp(request,"PROFILES HF|",12));
    int profiles_PIN     = (strlen(request)>=12 && !strncmp(request,"PROFILES PN|",12)); // this means that we need the PIN to do profile discovery
    LS_DEBUGT(profiles_a2dp, "profiles a2dp\n");
    LS_DEBUGT(profiles_headset, "profiles headset \n");
    LS_DEBUGT(profiles_protect, "profiles protect \n");
    LS_DEBUGT(profiles_PIN, "profiles PIN \n");

#ifdef SUPPORT_BRISTOL
    int profiles_hifi = (strlen(request)>=12 && !strncmp(request,"PROFILES HI|",12));
    LS_DEBUGT(profiles_hifi, "profiles hifi \n");
#endif  // SUPPORT_BRISTOL
    const char* bdaddr = &request[9];  // bdaddr starts at pos 9 by default

#ifdef SUPPORT_BRISTOL
    if (profiles_headset || profiles_protect || profiles_a2dp || profiles_hifi || profiles_PIN)
    {
      char* b = strstr(request, profiles_headset ? "HS|" : profiles_a2dp ? "A2DP|" : profiles_protect ? "HF|" : "HI|"); b+= (profiles_a2dp ? 5 : 3);
#else  // SUPPORT_BRISTOL
    if (profiles_headset || profiles_protect || profiles_a2dp || profiles_PIN)
    {
      char* b = strstr(request, profiles_headset ? "HS|" : profiles_a2dp ? "A2DP|" : "HF|"); b+= (profiles_a2dp ? 5 : 3);
#endif  // SUPPORT_BRISTOL
      char* e = b;
      while (*e && *e!='|') e++;
      int len=e-b; if (len>63) len=63;
      char pin[64]; strncpy(pin, b, len); pin[len]=0;
      bdaddr = e+(*e=='|' ? 1 : 0);  // bdaddr is specified after the pin

      // enable pairing temporarily, some headsets/phones need a pin during profile discovery (might need to pair first)
      system("killall -SIGUSR2 hcid");

      // update client pincode file
      char buf[128];
      sprintf(buf, "echo %s > %s", pin, BT_CLIENT_PINFILE);
      system(buf);

      LS_DEBUG("blueserver: Setting [%s] in [%s]\n", pin, BT_CLIENT_PINFILE);
    }

    scan_profiles_fd = fd;
    LS_DEBUG("blueserver: Scanning [%s]\n",bdaddr);
    scan_profiles_cmd_browse(bdaddr);
    writestr(fd, "END");

#ifdef SUPPORT_BRISTOL
    if (profiles_headset || profiles_protect || profiles_a2dp || profiles_hifi)
#else  // SUPPORT_BRISTOL
    if (profiles_headset || profiles_protect || profiles_a2dp)
#endif  // SUPPORT_BRISTOL
    {
      unlink(BT_CLIENT_PINFILE);  // remove client pinfile (so hcid uses default pin)
#ifndef  BT_HID_REMOTE
      system("killall -SIGUSR1 hcid");  // disable pairing again
#else
      LS_DEBUG("Pairing is NOT disabled, because of bluetooth remote control the app. must always be in pairing mode\n");
#endif
    }
  }
  else if (!strncmp(request,"ENABLE",6))
  {
    EnableBluetooth(1);
    writestr(fd, "END");
  }
  else if (!strncmp(request,"DISABLE",7))
  {
    EnableBluetooth(0);
    writestr(fd, "END");
  }
  else if (!strncmp(request,"MSRS ENABLE",11))
  {
    LS_DEBUG("blueserver: MSRS ENABLE\n");

    system("hciconfig hci0 lm accept");
    system("hciconfig hci0 lp rswitch,hold,sniff,park");
    writestr(fd, "END");
  }
  else if (!strncmp(request,"MSRS DISABLE",12))
  {
    LS_DEBUG("blueserver: MSRS DISABLE\n");

    system("hciconfig hci0 lm master");
    system("hciconfig hci0 lp hold,sniff,park");
    writestr(fd, "END");
  }
  else if (!strncmp(request,"ISCAN ENABLE",12)) /* enables visibility */
  {
    LS_DEBUG("blueserver: ISCAN ENABLE\n");
    int s;
    struct THCIDeviceRequirements dr;

    if ((s = socket(AF_BLUETOOTH, SOCK_RAW, BluetoothProtocolHCI)) != -1) {
      dr.iDeviceID  = 0;
      dr.iDeviceOptions = HciDeviceScanPage | HciDeviceScanInquiry;
      if (fcntl(s, F_SETFL, O_NONBLOCK) != -1) {
        if (ioctl(s, HCISETSCAN, (unsigned long)&dr) != -1) {
          LS_DEBUG("Iscan enabled");
        } else {
          LS_DEBUG("enabling Iscan failed");
        }
      } else {
        LS_DEBUG("enabling Iscan fcntl failed");
      }
      close(s);
    } else {
      LS_DEBUG("enabling Iscan socket failed");
    }
  }
  else if (!strncmp(request,"ISCAN DISABLE",13)) /* disables visibility */
  {
    LS_DEBUG("blueserver: ISCAN DISABLE\n");
    int s;
    struct THCIDeviceRequirements dr;

    if ((s = socket(AF_BLUETOOTH, SOCK_RAW, BluetoothProtocolHCI)) != -1) {
      if (fcntl(s, F_SETFL, O_NONBLOCK) != -1) {
        dr.iDeviceID  = 0;
        dr.iDeviceOptions = HciDeviceScanPage;
        if ((ioctl(s, HCISETSCAN, (unsigned long)&dr)) != -1) {
          LS_DEBUG("Iscan disabled");
        } else {
          LS_DEBUG("disabling Iscan ioctl failed");
        }
      } else {
        LS_DEBUG("disabling Iscan fcntl failed");
      }
      close(s);
    } else {
      LS_DEBUG("disabling Iscan socket failed");
    }
  }
  else if (!strncmp(request,"PAIRING ENABLE",14))
  {
    // "PAIRING ENABLE|<pin>"
    LS_DEBUG("blueserver: PAIRING ENABLE\n");

    // enable pairing temporarily, most devices need a pin during profile discovery or first connection
    system("killall -SIGUSR2 hcid");

    unlink(BT_CLIENT_PINFILE);
    if (strlen(request)>15 && !strncmp(request,"PAIRING ENABLE|",15))
    {
      char* b = strstr(request, "PAIRING ENABLE|"); b+= 15;
      char* e = b;
      while (*e && *e!='|') e++;
      int len=e-b; if (len>63) len=63;
      char pin[64]; strncpy(pin, b, len); pin[len]=0;

      // update client pincode file
      if (strlen(pin))
      {
        LS_DEBUG("blueserver: Writing PIN [%s] to %s\n", pin, BT_CLIENT_PINFILE);

        char buf[128];
        sprintf(buf, "echo %s > %s", pin, BT_CLIENT_PINFILE);
        system(buf);
      }
    }

    writestr(fd, "END");
  }
  else if (!strncmp(request,"PAIRING DISABLE",15))
  {
    LS_DEBUG("blueserver: PAIRING DISABLE\n");

    unlink(BT_CLIENT_PINFILE);  // remove client pinfile (so hcid uses default pin)

    // The check whether pairing should be disabled or not is moved to navcore
    system("killall -SIGUSR1 hcid");  // disable pairing
  }
  else
  {
    writestr(fd, "END");
  }
}

int main(int argc,char** argv)
{
  if (argc<2) exit(-1);
  char* the_address = argv[1];
#ifdef  BT_HID_REMOTE
  system("killall -SIGUSR2 hcid");
  LS_DEBUG("Pairing is enabled at startup, because of bluetooth remote control the app. must always be in pairing mode\n");
#endif
  return localserver(the_address,requesthandler);
}
